/*
 * Copyright (c) 2019, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

const COMPATIBLE_EVMID_REGEX = EVMID;
//const COMPATIBLE_EVMID_REGEX = /TMAG/;
const LATEST_FIRMWARE_VERSION = FIRMWARE_VERSION;

var myCodec;

// Global EVM object
var EVM = {
    id: {
        name: "",
        date: "",
        time: "",
        version: ""
    },
    preamble: { }
};

var Terminal = false;

/*
 * gc global variable provides access to GUI Composer infrastructure components and project information.
 * For more information, please see the Working with Javascript guide in the online help.
 */
var gc = gc || {};
gc.services = gc.services || {};


/* 
 * To see the console logs logged by this file, in the console view, enter the following:
 * gc.console.setCookie("pambCustomCodec.js", <value>) 
 * Where value: 0 = remove the cookie, 1 = only errors, 2 = errors and warnings, 3 = errors,warnings,info, 4 = errors,warnings,info,logs, 5=debug+all others
 * For more help on how to work with gc.console, please see the help text that is sent to the
 * console when the app starts.
 */
//if (!gc.designer) { gc.console.setCookie("pambCustomCodec.js", 5); }



CustomCodec = function () { };
    var initCodec = function () {

    gc.console.debug("pambCustomCodec.js", "The 'gc-databind-ready' event fired, building CustomCodec object.");

    if (!gc.designer) { gc.console.setCookie("pambCustomCodec.js", 5); }
    /**
     * This code implements the AbstractMessageBasedCodec interface.
     *
     * @constructor
     * @implements gc.databind.AbstractMessageBasedCodec
     */
    
    // IMPORTANT: Do NOT add any objects to the CustomCodec's protoype before calling the
    // gc.databind.AbstractMessageBasedCodec(), as they will get overwritten!
    
    //////////////////////// AbstractMessageBasedCodec APIs ////////////////////////
    
    // See additionals APIs and object provided by 'ti-core-databind\src\AbstractMessageBasedCodec.js'
    CustomCodec.prototype = new gc.databind.AbstractMessageBasedCodec("PAMB");
    // Objects: name, commandQueue, disconnect()
    
    // Overwriting the default 'connect' method in AbstractMessageBasedCodec...
    CustomCodec.prototype.connect = function(settings) {
        
        //
        // IMPORTANT: This method is called by the transport layer when attempting to connect to the serial COM port.
        //
        
        // Update the function scope for the AbstractMessageBasedCodec connect method
        gc.databind.AbstractMessageBasedCodec.prototype.connect.call(this);
        
        // Return a promise that will resolve after connection is complete
        return Q.Promise(function(resolve, reject) {
            
            // Set initial firmware state?
            this.setState("idle");                  // TODO: Make sure the firmware resets to idle state after connecting
            
            // Set connected flag
            this.connected = true;
            
            //
            // Send ID command, as soon as we are cleared to start sending commands
            //
            var identifyEVM;
            let sendIDcmd = function(event) {
                
                // Debugging...
                gc.console.debug("pambCustomCodec.js", "Handling 'pamb_clr_cmds' event");
                
                // Read ID of the EVM and make sure it is compatable with this GUI
                this.validateEvmId().then(function(evmId) {
                        
                    // Provide a list (or regular expression) of acceptable EVM names here...
                    // List example:
                        //const evmNameList = [ "ADS131M02", "ADS131M04", "ADS131M06", "ADS131M08" ];
                        //const match = (evmNameList.indexOf("EVM.id.name") > -1);
                    // RegEx example (https://developer.mozilla.org/en-US/docs/Web/JavaScript/Guide/Regular_Expressions):
                    const match = COMPATIBLE_EVMID_REGEX.test(evmId.name);
                    if (match) {
                        
                        // Check if an EVM firmware update is out-of-date
                        if (evmId.version < LATEST_FIRMWARE_VERSION) {
                            
                            // Release the USB interface
                            reject("Updating the EVM firmware...");
                            
                            // Send Invoke BSL command
                            this.encoder(EVM_INVOKE_BSL_COMMAND);
                            
                            alert('Firmware out of date. Will now update firmware.')
                            
                        } else {
                            
                            // TODO: Attempt to connect to the bulk device
                            
                            // Resolve connect promise
                            gc.console.debug("pambCustomCodec.js", "Resolving 'connect' promise");
                            resolve();
                            myCodec = this; 
                        }
                        
                    } else {
                        // Reject connect promise -> This will call the disconnect() method
                        gc.console.debug("pambCustomCodec.js", "Rejecting 'connect' promise");
                        reject("This application is not compatabile with the " + EVM.id.name + ". Please \
                        try a different EVM or GUI Composer application!");
                    }
                    
                }.bind(this)).catch(function(err) {
                    // TODO: handle error
                });
                
                // Remove the event listener since this event will only execute once
                document.removeEventListener('pamb_clr_cmds', identifyEVM);
            };
            identifyEVM = sendIDcmd.bind(this);
            
            // Add event listener for 'pamb_clr_cmds' event
            document.addEventListener('pamb_clr_cmds', identifyEVM);
            
            // TODO: Remove?
            identifyEVM();
            
        }.bind(this));
    };
    
    // Overwriting the default 'disconnect' method in AbstractMessageBasedCodec...
    CustomCodec.prototype.disconnect = function() {
        
        gc.databind.AbstractMessageBasedCodec.prototype.disconnect.call(this);
        
        // Set connection flag
        this.connected = false;
        
        // Clear all pending commands and deferred writeValue and readValue promises
        this.pendingSendCmds = [];
        this.commandQueue.clearAll();
        
        // Call close function()
        closeBulkDevice(false); // 'false' indicates that this is not in response to a detach event
        
        // NOTE: Remove all event listeners you added in the connect function here,
        // unless they were only called once and removed from within the callback function.
        // See reference: https://stackoverflow.com/questions/23805878/why-call-removeeventlistener-inside-addeventlistener-callback
    };
    
    // Overwriting the default 'isStillConnected' method in AbstractMessageBasedCodec...
    // This function is periodically called by the internal state machine to check if hardware is connected.
    CustomCodec.prototype.isStillConnected = function() {
        
        // Resolve promise based on our internal state flag
        if (this.connected) {
            return Q.resolve(this.connected);
        } else {
            return Q.reject(this.connected);
        }
        
        // (OPTIONAL) Ping the hardware and check for a response
    };
    
    
    /////////////////////////// IPollingPacketCodec APIs ///////////////////////////
    
    // NOTE: AbstractMessageBasedCodec also implements the IPollingPacketCodec
    // CustomCodec.prototype = new gc.databind.IPollingPacketCodec();
    
    
    /**
     * Encodes data into a packet for sending to the target device.
     *
     * @param target - function to call to pass on the encoded data towards the target
     * @param jsonObj - The data object from the GUI that is to be encoded into a packet for sending.
     */
    CustomCodec.prototype.encode = function(target, cmdString) {
        
        // Do not send commands while EVM is busy.
        if (this.evmState === "busy") {
            
            // Enque command
            // this.pendingSendCmds.push({
            //     target: target,
            //     cmd: cmdString
            // });
            // gc.console.debug("pambCustomCodec.js", "Enquing command '" + cmdString + "' (" + this.pendingSendCmds.length + " cmds pending)");
            // return;
            this.enqueueCmd(target, cmdString);
            return;
            
        } else if ((this.evmState === "collecting") && (cmdString !== "collectstop")) {
            
            // Command not allowed while collecting data
            return;
        }
        
        // Send command...
        this.strTxBuffer = cmdString; // We will use this buffer for recalling the last command sent to the device
        if (this.strTxBuffer.length > 0) {
            
            // Update EVM state to busy
            this.setState("busy");
            
            // Debugging...
            let msg = "Sending command '" + this.strTxBuffer + "'";
            if (Terminal) { Terminal.writeTextArea("\n" + msg + "\n"); }
            
            // Log message (NOTE: Separator is used to make the log easier to read)
            gc.console.info("pambCustomCodec.js", "------------------------------------------");
            gc.console.info("pambCustomCodec.js", msg);
            
            try {
                target(this.strTxBuffer);
            } catch (ex) {
                gc.console.error("pambCustomCodec.js", "CustomCodec.encode: exception=" + ex);
            }
        }
        
        // Set 3 second timeout in case firmware does not respond to command
        // NOTE: The timer will be cleared when the firmware acknowledges the command.
        // this.timerID = setTimeout(function() {
        //     this.connected = false;
        //     alert("Timeout: Firmware failed to acknowledge '" + this.strTxBuffer + "' command within 3 seconds!");
        // }.bind(this), 3000);
    };
    
    
    /**
     * Decodes packets from the target device into data objects.  One object for each packet of data.
     * Unless you have chained your custom codec with the CR codec, the packet data is not framed,
     * so this method is responsible for decoding partial packets as well as multiple packets.  Partial packets
     * should not be returned, but deferred until the next decode method call with more raw data to process.  As a
     * result, this method may return 0 or more packets of data on each invocation.
     *
     * @param target -  function to call to pass on a JSON object containing the decoded values towards the GUI
     * @param data  - The raw data received that needs to be decoded from packets into objects.
     * @return - true if connected (i.e. valid data received), false if not connected
     */
    CustomCodec.prototype.decode = function(target, data) {
        
        // Convert data to ASCII
        var strRxData = data.map(function(value) {
            return String.fromCharCode(value);
        }).join("");
        
        // Add RX data to strRxBuffer and look for a newline character...
        this.strRxBuffer = this.strRxBuffer + strRxData;
        while (this.strRxBuffer.indexOf('\n') !== -1) {
            
            // Parse JSON message from strRxBuffer (message must always end with newline character!)
            var strMessage = this.strRxBuffer.substring(0, this.strRxBuffer.indexOf('\n') + 1);
            gc.console.info("pambCustomCodec.js", "Received response " + strMessage);
                
            // Remove JSON message from strRxBuffer
            this.strRxBuffer = this.strRxBuffer.substring(strMessage.length);
                
            // Check for a message to parse
            if (!strMessage) { break; }
                
            // Update Terminal text area
            if (Terminal) { Terminal.writeTextArea(strMessage); }
                
            // Try to parse message as a JSON object
            try {
                
                // Parse JSON string to object
                var jsonObj = JSON.parse(strMessage.trim());
                
                //
                // Look for predefined JSON properties to decide what action(s) to take...
                //
                
                // EVM firmware is acknowledging a recieved command...
                if (jsonObj.hasOwnProperty("acknowledge")) {
                    
                    // Clear firmware timeout
                    clearTimeout(this.timerID);
                    
                    // Get acknowledged command string
                    let cmd_ack = jsonObj.acknowledge;
                    
                    // Update Terminal
                    if (Terminal) { Terminal.acknowledge(cmd_ack); }
                    
                    // Check if this response does not match the last command sent to the EVM...
                    if (cmd_ack !== this.strTxBuffer) {
                        gc.console.error("pambCustomCodec.js", "The EVM's acknowledged command ('" + cmd_ack + "') does not match the last command sent ('" + this.strTxBuffer + "')");
                    }
                    
                    // NOTE: In the case where writeValue() is waiting on a promise, you may want to resolve that promise here:
                    // this.resolveCmdPromise(/^wreg/, undefined, false);
                    
                }
                
                // Log firmware message to console...
                if (jsonObj.hasOwnProperty("console")) {
                    
                    // Get message
                    let msg = jsonObj.console;
                    
                    // Check for message prefix of "Error", "Warn", or "Info" to direct log output type
                    let msgType;
                    if (msg.substr(0, 5).toLowerCase() === "error") {
                        msgType = "error";
                        gc.console.error("pambCustomCodec.js", msg);
                        
                    } else if (msg.substr(0, 5).toLowerCase() === "warn") {
                        msgType = "warn";
                        gc.console.warning("pambCustomCodec.js", msg);
                        
                    } else {
                        msgType = "info";
                        gc.console.log("pambCustomCodec.js", msg);
                    }
                    
                    // Output message to application's eventlog widget
                    if (this.eventLog) { this.eventLog.log(msgType, msg); }
                }
                
                // Firmware is reporting an error...
                if (jsonObj.hasOwnProperty("error")) {
                    // TODO: build list of possible error numbers
                }
                
                // (OPTIONAL) EVM device mode has changed...
                if (jsonObj.hasOwnProperty("device_modes")) {
                    // Extract mode information and store it for reference...
                }
                
                // EVM ID
                if (jsonObj.hasOwnProperty("id")) {
                    
                    // Update the EVM object
                    EVM.id = jsonObj.id;
                    EVM.id.version = this.firmwareVersionNumber(jsonObj.id);
                    
                    // Print info to event log
                    if (this.eventLog) { this.eventLog.log("EVM identified as '" + EVM.id.name + "'. Firmware version = '" +  EVM.id.version + "'."); }
                    
                    // Print info to gc console
                    gc.console.log("pambCustomCodec.js", "EVM identified as '" + EVM.id.name + "'. Firmware version = '" +  EVM.id.version + "' (" + EVM.id.date + " " + EVM.id.time + ")");
                    
                    // Resolve promise (if any)
                    this.resolveCmdPromise(/^id/, EVM.id, false);
                    
                }
                
                // Information about the binary data formatting
                // IMPORTANT: THIS WILL NEED TO BE UPDATED FOR EACH EVM GUI!
                // TODO: Move to a collect function
                if (jsonObj.hasOwnProperty("preamble")) {
                    
                    //
                    // Build FrameDefinitions array from current register map settings
                    //
                    
                    // Get number of channels...
                    let idVal = this.getRegisterValue(0x00);   // 0x00 = address of ID reigster
                    channelCount = (idVal & 0x0F00) >> 8;
                    
                    // Get data byte length
                    let modeVal = this.getRegisterValue(0x02);   // 0x02 = address of MODE reigster
                    let wlengthVal = (modeVal & 0x0300) >> 8;
                    bytesPerWord = (wlengthVal < 3) ? wlengthVal+2 : 4; // Convert field value to number of bytes, per device datasheet
                    
                    //
                    // FRAME 0
                    //
                    FrameDefinitions[0x00] = [];    // NOTE: When building a 2D array, we need to initialize each frame separately.
                    
                    // Frame ID (required) 
                    FrameDefinitions[0x00].push(new CreateChannel("frameId", 1, false, null));
                    
                    // STATUS
                    FrameDefinitions[0x00].push(new CreateChannel("status", bytesPerWord, false, statusArray));
                    
                    // CHANNELS 
                    for (let i = 0 ; i < channelCount; i++) {
                        // NOTE: We create a channel data placeholder for all channels, even if they are disabled
                        channelData[i] = [];
                        FrameDefinitions[0x00].push(new CreateChannel("channel" + i, bytesPerWord, true, channelData[i]));
                    }
                    
                    // CRC
                    FrameDefinitions[0x00].push(new CreateChannel("crc", bytesPerWord, false, crcArray));
                }
                
                // Register data
                if (jsonObj.hasOwnProperty("register")) {
                    
                    // Get register read value
                    let readValue = jsonObj.register.value;
                    
                    // Check if we are resolving a promise...
                    let resolved = this.resolveCmdPromise(/^rreg/, +readValue, false);
                    if (!resolved) {
                        
                        // Update the register model directly
                        this.setRegisterValue(jsonObj.register.address, readValue);
                        
                    }
                }
                
                // EVM state (tells us if EVM is 'idle' and ready to recieve commands)
                if (jsonObj.hasOwnProperty("evm_state")) {
                    this.setState(jsonObj.evm_state);
                }
                
                // Send the received and decoded javascript object to the application where you can use it to bind to widget properties
                target(jsonObj);
                
            } catch (ex) {
                gc.console.error("pambCustomCodec.js", "decode (from target): strMessage='" + strMessage + "', exception=" + ex);
            }
        }
        
        // Return true to set the 'Hardware Connected' status in the status bar
        return this.connected;
    };
    
    
    
    /////////////////////////
    // Register model APIs //
    /////////////////////////
    
    /*
     * Confluence: https://confluence.itg.ti.com/pages/viewpage.action?pageId=173550282
     * For an example of a codec that accesses symbols using readValue/writeValue, please see components/ti-widget-registers/demo/MotorwareDemo_F28069M/customCodec_0.js
     */
    
    // Called by register model to read a device register...
    CustomCodec.prototype.readValue = function(regInfo) {
        
        const addressString = (regInfo.addr).toString(16);
        
        // (OPTIONAL) Check for valid register address and reject promise if invalid
        if (regInfo.addr > 0x3F) {
            gc.console.warning("pambCustomCodec.js", "REJECTING readValue : " + regInfo.addr);
            return Q.reject("bad register addr");
        }
        
        // (OPTIONAL) Ignore register readback under certain conditions
        const ignoreAddress = 0x00;
        const ignoreMask = 0x70;
        const ignoreValue = 0x60;
        if(this.blockReadback === true) {
            this.blockReadback = false;
            //these next two if statements are technically redundant from the write section
            if (regInfo.addr === ignoreAddress) { 
                const currentRegValue = this.getRegisterValue(ignoreAddress); 
                if ((currentRegValue & ignoreMask) === ignoreValue) {
                        this.setRegisterValue(ignoreAddress, ignoreValue); //May not need, clears itself later
                        return ignoreValue; // Was currentRegValue, using ingnore value is device specific 
                }
            }
        }
        
        this.inDeepSleep = false;
        this.blockReadback = false;
        
        // Build command string
        const cmd = READ_REGISTER_COMMAND + ' ' + addressString;
        
        
        // Add the pending command to the internal commandQueue APIs provided by 
        // ti-core-databind\src\AbstractMessageBasedCodec.js
//        let cmd = READ_REGISTER_COMMAND + ' ' + (regInfo.addr).toString(16);
        let promise = this.addCommand(cmd, this.pendingCmdSeqNum++);
        
        // Debugging...
        gc.console.log("pambCustomCodec.js", "readValue : " + cmd);
        
        // Send read register command
        this.encoder(cmd);
        
        return promise;
    };
    
    // Called by register model to write register data...
    CustomCodec.prototype.writeValue = function(regInfo, value) {
        
        // Build the write register command string
        let cmd = WRITE_REGISTER_COMMAND + ' ' + (regInfo.addr).toString(16) + ' ' + value.toString(16);
        
        // Debugging...
        gc.console.log("pambCustomCodec.js", "writeValue : " + cmd);
        
        this.blockReadback = false;
        
        
        const ignoreAddress = 0x00;
        const ignoreMask = 0x70;
        const ignoreValue = 0x60; //Deep Sleep Mode
        if (regInfo.addr === ignoreAddress) {
            if ((value & ignoreMask) === ignoreValue) {
                if (this.inDeepSleep === true) {
                    this.blockReadback = false; //Already in Deep Sleep, so we are now waking up... (may be unintentional by user/GUI)
                    this.inDeepSleep = false;
                }
                else {
                    this.blockReadback = true; 
                    this.inDeepSleep = true;
                }
            }
            else {
                this.inDeepSleep = false;
            }
        }  
        else {
            this.inDeepSleep = false;
        }

        
        // Send write register command to queue the command locally
        this.encoder(cmd);
        
        // Return a completed promise (in this case with undefined value since writeValue does not look at the result, it just waits for the promise to complete.
        return Q();
        
        // NOTE: There is no need to delay the completion of the writeValue promise,
        // but if for some reason you need to, then here is an alternate implementation.
        // Add the pending command to the internal commandQueue APIs provided by 
        // ti-core-databind\src\AbstractMessageBasedCodec.js and get a promise
        // let promise = this.addCommand(cmd, this.pendingCmdSeqNum++);
        // gc.console.log("pambCustomCodec.js", "writeValue: this.commandQueue.nextCmd.command = " + this.commandQueue.first.next.command);
        // return promise;
    };
    
    // Lookup function to get the fully qualified register name from the register address.
    CustomCodec.prototype.findRegister = function(regAddrs) {
        
        let result; // Default return value: undefined
        
        // If no cached info, then we cannot perform the lookup
        if (!this.cachedRegBlocks) { return result; }
        
        // Loop through register blocks looking for matching register address
        let found = [];
        for (let i = 0; i < this.cachedRegBlocks.length; i++) {
            
            // Filter array looking for a matching register address
            found = this.cachedRegBlocks[i].registers.filter(r => r.addr === +regAddrs);
            
            // Stop as soon as we find a match
            if (found.length > 0) {
                result = found[0];
                break;
            }
        }
        return result;
    };
    
    // Lookup function to get the current register value from a specified register address.
    CustomCodec.prototype.getRegisterValue = function(regAddrs) {
        
        // Get register object which contains register info
        let regObj = this.findRegister(regAddrs);
                    
        // Check if the register object exists (in case it is not defined in JSON file, for example)
        if (regObj) {
    
            // Get register value...
            let regBind = gc.databind.registry.getBinding(this.registerModel.id + '.' + regObj.fullyQualifiedName);
            return regBind.getValue();
            
        } else {
            gc.console.error("pambCustomCodec.js", "getRegisterValue: Could not locate register with address: " + regAddrs);
            return -1;
        }
    };
    
    // Lookup function to set a register value from a specified register address.
    CustomCodec.prototype.setRegisterValue = function(regAddrs, value) {
        
        // Get register object which contains register info
        let regObj = this.findRegister(regAddrs);
        
        // Check if the register object exists (in case it is not defined in JSON file, for example)
        if (regObj) {
            
            // Update the register model with the new register data...
            let regBind = gc.databind.registry.getBinding(this.registerModel.id + '.' + regObj.fullyQualifiedName);
            regBind.updateValue(value);
            
        } else {
            gc.console.error("pambCustomCodec.js", "setRegisterValue: Could not locate register with address: " + regAddrs);
        }
    };
    
    
    //////////////////////// Custom Properties ////////////////////////
    
    // TODO: Define additional commands here...
    const WRITE_REGISTER_COMMAND = 'wreg';
    const READ_REGISTER_COMMAND = 'rreg';
    const EVM_ID_COMMAND = 'id';
    const EVM_INVOKE_BSL_COMMAND = 'bsl';
    
    
    // RX and TX buffers
    CustomCodec.prototype.strRxBuffer = "";
    CustomCodec.prototype.strTxBuffer = "";
    
    // Command queueing
    CustomCodec.prototype.evmState = "";
    CustomCodec.prototype.pendingSendCmds = [];
    CustomCodec.prototype.pendingCmdSeqNum = 0;
    CustomCodec.prototype.cmdSeqNum = 0;
    
    // Internal state flags
    CustomCodec.prototype.initComplete = false;
    CustomCodec.prototype.connected = false;
    
    // References to important objects
    CustomCodec.prototype.registerModel     = undefined;
    CustomCodec.prototype.cachedRegBlocks   = undefined;  // Placeholder for saving register model information
    CustomCodec.prototype.eventLog          = undefined;
    CustomCodec.prototype.timerID           = undefined;  // Timer ID for firmware timeout
    
    CustomCodec.prototype.blockReadback         = false;
    CustomCodec.prototype.inDeepSleep         = false;
    
    // Save reference to register model, and cache the register data
    CustomCodec.prototype.registerModel = document.querySelector('ti-model-register');
    if (CustomCodec.prototype.registerModel) {
        // Cache register block information
        let deviceIndex = 0;
        CustomCodec.prototype.registerModel.getRegisterViewInformation(deviceIndex).then(function(deviceObj) {
            CustomCodec.prototype.cachedRegBlocks = deviceObj.regblocks;
        });
    }
    
    // Save reference to eventlog
    CustomCodec.prototype.eventLog = document.querySelector('ti-widget-eventlog-view');
    
    // Set initialization complete flag
    CustomCodec.prototype.initComplete = true;
    
    ///////////////////////// CustomCodec Helper functions /////////////////////////
    
    // Call this function after the 'dom-change' event...
    CustomCodec.prototype.init = function() {
        
    };
    
    
    // Called by register model to read a device register...
    CustomCodec.prototype.validateEvmId = function() {
        
        // Add the pending command to the internal commandQueue
        let cmd = EVM_ID_COMMAND;
        let promise = this.addCommand(cmd, this.pendingCmdSeqNum++);
        
        // Debugging...
        gc.console.log("pambCustomCodec.js", "Validating EVM identity...");
        
        // Read ID of the EVM and make sure it matches
        this.encoder(EVM_ID_COMMAND);
        
        // TODO: Create a timeout for the response that will reject thie prmoise if we do not recieve a resposne in a timely manner,
        // For example, we may connect to an incorrect serial com port and the connected device may not respond to this command.
        
        return promise;
    };
    
    // Update EVM state variable
    CustomCodec.prototype.setState = function(state) {
        
        //  Check if state is valid...
        const re = new RegExp('^idle$|^busy$|^collecting$');
        const isValid = re.test(state);
        if (!isValid) {
            gc.console.error("pambCustomCodec.js", "Invalid EVM state: '" + state + "' in CustomCodec!");
        }
        
        // Save previous state
        let prevState = this.state;
        
        // Update internal state variable and indicator
        this.evmState = state;
        if (Terminal) { Terminal.setState(state); }
        
        // Handle state transistions...
        switch (state) {
            
            case "idle":
                
                if (prevState == "collecting") {
                    
                    // Analyze new data
                    channelData[0];     // TODO: Implement
                    
                    
                }
                
                
                // Dequeue next command (if any)...
                // NOTE: Make sure to update the internal state variable before
                // calling this function, or else the dequeuing method will fail!
                this.dequeueCmd();
                
                break;
                
            case "busy":
                break;
                
            case "collecting":
                
                // Clear old data
                Parser.clear(FrameDefinitions);
                
                break;
        }
    };
    
    // TODO: Write a function that combines the functionality of encode & decode
    // to issue a command to the EVM and return the JSON response (for scripting)
    
    // Enqueue command
    CustomCodec.prototype.enqueueCmd = function(target, cmdString) {
        
        // Enque command
        this.pendingSendCmds.push({
            target: target,
            cmd: cmdString,
            //promise: this.addCommand(cmdString, this.pendingCmdSeqNum++) // NOTE: We may not want to create a promise with each command...
        });
        gc.console.debug("pambCustomCodec.js", "Enquing command '" + cmdString + "' (" + this.pendingSendCmds.length + " cmds pending)");
        //return promise;
    };
    
    // Dequeue command
    CustomCodec.prototype.dequeueCmd = function() {
        
        // Check if we have any pending commands to dequeue
        if (this.pendingSendCmds.length > 0) {
            let nextCmd = this.pendingSendCmds.shift();
            gc.console.debug("pambCustomCodec.js", "Dequing command '" + nextCmd.cmd + "' (" + this.pendingSendCmds.length + " cmds pending)");
            this.encode(nextCmd.target, nextCmd.cmd);
        }
    };
    
    // Resolve command promise
    CustomCodec.prototype.resolveCmdPromise = function(regex, response, error) {
        
        // Check if we are resolving a promise...
        let pendingCmd = this.commandQueue.first.next;
        if (!pendingCmd || !pendingCmd.command) { return false; }
        
        if (regex.test(pendingCmd.command)) {
            
            let msg = "Resolving '" + pendingCmd.command + "' command promise [seqNo: " + pendingCmd.seqNo + ", isError: " + error + "] with " + JSON.stringify(response);
            gc.console.log("pambCustomCodec.js", msg);
            
            // Resolve promise based on match result
            this.addResponse(response, pendingCmd.command, pendingCmd.seqNo, error);
            this.cmdSeqNum++;
            
            if (!pendingCmd.next) {
                gc.console.debug("pambCustomCodec.js", "The commandQueue is now empty");
            }
            
            return true;    // Indicate to caller that we resolved a promise
        } else {
            return false;   // Indicate to caller that we did not resolve a promise
        }
    }
    
    // Build firmware version number from compiler's date/time string
    CustomCodec.prototype.firmwareVersionNumber = function(idObj) {
        
        let date = idObj.date.split(" "); // Ex: 'Feb','','7','2019'
        date = date.filter(entry => entry.trim() !== ''); // Remove any null entries that occur on single digit days
        
        let time = idObj.time.split(":"); // Ex: '16','48','59'
        
        // REFERENCE: http://infocenter.arm.com/help/index.jsp?topic=/com.arm.doc.faqs/ka9555.html
        const month_names = ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        
        // Get index of matching month abbreviaiton
        let month_index = 1 + month_names.findIndex(function(element) {
            return (element === date[0]);
        });
        
        // Convert values to strings, with length 2 (except for year)
        let year = date[2];
        let month = MyUtil.pad(month_index, 2); //= (month_index < 10) ? ('0' + month_index) : month_index;
        let day = (date[1].length < 2) ? ('0' + date[1]) : date[1];
        let hour = (time[0].length < 2) ? ('0' + time[0]) : time[0];
        let min = (time[1].length < 2) ? ('0' + time[1]) : time[1];
        let sec = (time[2].length < 2) ? ('0' + time[2]) : time[2];
        
        // Concatenate strings
        let versionString = ("" + year + month + day + hour + min + sec);
        return versionString;
    };
    
    CustomCodec.prototype.collectData = function() {
        // TODO: Determine how to stream or request a specific number of samples..
    };
    
    
    ///////////////////////////// Register CustomCodec /////////////////////////////
    
    // registerCustomCodec has to be called right at startup in order for the codec to be recognized and used
    gc.console.log("pambCustomCodec.js", "CustomCodec: about to register...");
    
    // You can 'chain' codecs together to make it easier to work with the data
    // coming from the device or going to the device.
    // The following codecs are provided:
    //   CR  - use this if the target data uses /n as a delimiter for framing.
    //         The CR codec will take care of buffering the incoming data so that
    //         your codec will only receive complete 'ready to use' packets
    //  If your target uses \n delimited strings, set baseCodecs = "CR"
    //  If your target uses some other format for data (e.g. binary data or custom strings) leave baseCodecs = "undefined"
    var baseCodecs;
    gc.databind.registerCustomCodec("custom", CustomCodec, baseCodecs);

    gc.console.debug("pambCustomCodec.js", "CustomCodec object registered.");
    
} // end 'gc-databind-ready' eventlistener function

var templateObj = document.querySelector('#template_obj');
if (templateObj) {
    initCodec();
} else {
    document.addEventListener('WebComponentsReady', initCodec);
}    

/*
*  Boilerplate code for working with components in the application gist


var initComplete = false;
var templateObj;
var modelObj;

// Wait for DOMContentLoaded event before trying to access the application template
var init = function() {
    
    templateObj = document.querySelector('#template_obj');

    // Wait for the template to fire a dom-change event to indicate that it has been 'stamped'
    // before trying to access components in the application.
	templateObj.addEventListener('dom-change', function() {
	    
	    if (initComplete) { return; }
	    
	    this.async(function() {
    	    initComplete = true;
    	    gc.console.debug("terminal.js", "Application template has been stamped.");
  	        
  	        // Now that the template has been stamped, you can use 'automatic node finding' $ syntax to access widgets.
  	        // e.g. to access a widget with an id of 'widget_id' you can use templateObj.$.widgetId

            // Save reference to model object (only available after initialization has completed)
            modelObj = document.querySelector('ti-model-streaming');

            // Clear text area button event
            var clear_button = document.querySelector(TERMINAL_CLEAR_BUTTON);   
            clear_button.addEventListener("click", function(event) {
                Terminal.clearTextArea();
            });

            // TODO: Create valueChanged event to validate command inputs
            // Reference: https://stackoverflow.com/questions/6039565/character-limit-on-textbox
            
            // TODO: Get command list from customCodec and provide drop-down suggestions

            // TX button event
            var tx_button = document.querySelector(TERMINAL_TX_BUTTON);   
            tx_button.addEventListener("click", function(event) {
                var commandString = document.querySelector(TERMINAL_TXSTRING_TEXTBOX).value;
                // TODO: Limit text string length to CMD_BUF_SIZE, too long will lockup firmware
                modelObj.sendValue(commandString);
            });
        }, 1);

	});
};

templateObj = document.querySelector('#template_obj');
if (templateObj) {
    init();
} else {
    document.addEventListener('DOMContentLoaded', init.bind(this));
}
*/

///////////////////////// Binary data parsing classes //////////////////////////

// Array buffer methods for converting binary to integer types
var ArrayBufferCursor = (function() {
    "use strict";
    
    // REFERENCE: https://riptutorial.com/javascript/example/32392/iterating-through-an-arraybuffer
    
    // Constructor
    var ArrayBufferCursor = function(arrayBuffer) {
        this.dataview = new DataView(arrayBuffer, 0);
        this.size = arrayBuffer.byteLength;
        this.index = 0;
        this.isLittleEndian = false;
    };
    
    ArrayBufferCursor.prototype.next = function(type) {
        let result;
        let unsignedINT24;
        switch(type) {
            
            // NOTE: For 8-bit data types, the array buffer index is being incremented inline...
            
            case 'Int8':
                result = this.dataview.getInt8(this.index++, this.isLittleEndian);
                return result;
            
            case 'Uint8':
                result = this.dataview.getUint8(this.index++, this.isLittleEndian);
                return result;
                
            case 'Int16':
                result = this.dataview.getInt16(this.index, this.isLittleEndian);
                this.index += 2;
                return result;
                
            case 'Uint16':
                result = this.dataview.getUint16(this.index, this.isLittleEndian);
                this.index += 2;
                return result;
                
            case 'Int24':
                unsignedINT24 = this.dataview.getUint8(this.index++, this.isLittleEndian);
                unsignedINT24 = (unsignedINT24 << 8) + this.dataview.getUint8(this.index++, this.isLittleEndian);
                unsignedINT24 = (unsignedINT24 << 8) + this.dataview.getUint8(this.index++, this.isLittleEndian);
                return ((unsignedINT24 & 0x800000) ? (unsignedINT24 - Math.pow(2, 24)) : unsignedINT24);
                
            case 'Uint24':
                unsignedINT24 = this.dataview.getUint8(this.index++, this.isLittleEndian);
                unsignedINT24 = (unsignedINT24 << 8) + this.dataview.getUint8(this.index++, this.isLittleEndian);
                unsignedINT24 = (unsignedINT24 << 8) + this.dataview.getUint8(this.index++, this.isLittleEndian);
                return unsignedINT24;
                
            case 'Int32':
                result = this.dataview.getInt32(this.index, this.isLittleEndian);
                this.index += 4;
                return result;
                
            case 'Uint32':
                result = this.dataview.getUint32(this.index, this.isLittleEndian);
                this.index += 4;
                return result;
                
            // UNSUPPORTED TYPES...
            // case 'Float32':
            //     result = this.dataview.getFloat32(this.index, this.isLittleEndian);
            //     this.index += 4;
            //     return result;
            
            // case 'Float64':
            //     result = this.dataview.getFloat64(this.index, this.isLittleEndian);
            //     this.index += 8;
            //     return result;
            
            default:
                throw new Error("Unknown datatype");
        }
    };
    
    ArrayBufferCursor.prototype.hasNext = function(typeString) {
        if (typeString) {
            
            // Get number of bytes from type string (e.g. 'Uint24' => 24 / 8 = 3)
            var numberBytes = (typeString.match(/\d+/)) / 8;
            
            // Check if there are enough bytes to extract this data type
            return (this.size - this.index) >= numberBytes;
            
        } else {
            
            // Check if there is at least 1 byte in arraybuffer
            return this.index < this.size;
        }
    };

    return ArrayBufferCursor;
})();

// Binary data parser for sorting data
var BinaryDataParser = (function() {
    "use strict";
    
    // Constructor
    var BinaryDataParser = function() {
        this.frameId = 0;
        this.channelCounter = 0;
    };
    
    BinaryDataParser.prototype.getCursor = function(data) {
        var ArrayObj = new Uint8Array(data);
        return new ArrayBufferCursor(ArrayObj.buffer);
    };
    
    BinaryDataParser.prototype.parse = function(binaryDataArray, dataFrames) {
        
        // Get cursor
        var cursor = this.getCursor(binaryDataArray);
        
        // Get current frame
        var currentFrame = dataFrames[this.frameId];
        if (!currentFrame) {
            gc.console.error("pambCustomCodec.js", "Data parsing misalignment: Invalid frame id: " + this.frameId + "!");
        }
        
        // Get current channel
        var channel = currentFrame[this.channelCounter];
        if (!channel) {
            gc.console.error("pambCustomCodec.js", "Data parsing error: Invalid channel: " + this.channelCounter + "!");
        }
        
        // Parse data until we have an insufficent number of bytes to continue...
        while(cursor.hasNext(channel.type))
        {
            // Parse data into channel's data array
            if (channel.dataArrayPtr) {
                var value = cursor.next(channel.type);
                channel.dataArrayPtr.push(value);
                //gc.console.log("pambCustomCodec.js", channel.name + " " + channel.type + " " + value);
                
            } else {
                
                // IMPORTANT: We are assuming that only the frameID channel will have a null 'data' object!!!
                this.frameId = cursor.next(channel.type);
                if (this.frameId >= dataFrames.length) {
                    gc.console.error("pambCustomCodec.js", "Data parsing misalignment: Invalid frame id: " + this.frameId + "!");
                }
                currentFrame = dataFrames[this.frameId];
                //gc.console.log("pambCustomCodec.js", "frame ID: " + this.frameId);
            }
            this.channelCounter = (this.channelCounter + 1) % currentFrame.length;
            channel = currentFrame[this.channelCounter];
        }
        
        // Keep the residual data, and removed parsed data from array
        // NOTE: http://www.tothenew.com/blog/javascript-splice-vs-slice/
        binaryDataArray.splice(0, cursor.index);
    };
    
    // Call this method before beginning a new data collections
    BinaryDataParser.prototype.clear = function(dataFrames) {
        
        // Reset internal variaables
        this.frameId = 0;
        this.channelCounter = 0;
        
        // Clear all channel data arrays
        dataFrames.forEach((frame) => {
            frame.forEach((channel) => {
                if (channel.dataArrayPtr) {
                    channel.dataArrayPtr.length = 0;   // Clear array by setting it's length to zero
                }
            });
        });
    };

    return BinaryDataParser;
    
})();

// Channel class used by BinaryDataParser and FrameDefinitions for sorting data into various "channels" 
var CreateChannel = function(nameStr, byteLen, isSigned, dataArray) {
    this.name = nameStr;
    this.size = byteLen;
    this.type = (isSigned ? "Int" : "Uint") + (8*byteLen);  // NOTE: This currently does not support float types!
    this.dataArrayPtr = dataArray;
};

// Create an instance of the BinaryDataParser class (TODO: Move this somewhere else)
var Parser = new BinaryDataParser();


//
// Binary data frame definitions
//

// Channel parameters
var channelCount = 4;   // Default value; may change at runtime 
var bytesPerWord = 3;   // Default value; may change at runtime 

// Data placeholders...  // TODO: Where to attach these arrays, model or codec?
var configArray = [];
var channelData = [];
var crcArray = [];

var FrameDefinitions = [];  // NOTE: This array may need to be dynamically built if the data formatting is configurable by the user!

// FRAME 0
FrameDefinitions[0x00] = [];    // NOTE: When building a 2D array, we need to initialize each frame separately.

    // Frame ID (required) 
    FrameDefinitions[0x00].push(new CreateChannel("frameId", 1, false, null));
    
    // STATUS
    FrameDefinitions[0x00].push(new CreateChannel("Config", 2, false, configArray));
    
    // // CHANNELS 
    // for (let i = 0 ; i < channelCount; i++) {
    //     // NOTE: We create a channel data placeholder for all channels, even if they are disabled
    //     channelData[i] = [];
    //     FrameDefinitions[0x00].push(new CreateChannel("channel" + i, bytesPerWord, true, channelData[i]));
    // }
    
    // // CRC
    // FrameDefinitions[0x00].push(new CreateChannel("crc", 2, false, crcArray));
    
// FRAME 1
FrameDefinitions[0x01] = [];    // NOTE: When building a 2D array, we need to initialize each frame separately.

    // Frame ID (required) 
    FrameDefinitions[0x01].push(new CreateChannel("frameId", 1, false, null));
    
    // ADC Config
    FrameDefinitions[0x01].push(new CreateChannel("ADC Config", 2, false, configArray));
    
